/******************************************************************************
    (c) 1998      P.J. Caulfield          patrick@pandh.demon.co.uk
                  E.M. Serrat             emserrat@geocities.com
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 ******************************************************************************
 */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netdnet/dn.h>
#include <netdnet/dnetdb.h>
#include <unistd.h>


struct	sockaddr_dn		sockaddr;
struct	accessdata_dn		accessdata;
static  struct	nodeent        *binadr;
static	char			dirname[80];
static	unsigned char		buf[32760];
static	int			buflen,bufptr;
static  char                   *lasterror;

/* DECnet phase IV limits */
#define MAX_NODE      6
#define MAX_USER     12
#define MAX_PASSWORD 12
#define MAX_ACCOUNT  12
#ifndef FALSE
#define TRUE 1
#define FALSE 0
#endif

static int parse(char *);

/*-------------------------------------------------------------------------*/
static void usage(void)
{
    printf("\nUSAGE: dndel [OPTIONS] 'node\"user password\"::filespec'\n\n");
    printf("NOTE: The VMS filename really should be in single quotes to\n");
    printf("      protect it from the shell\n");

    printf("\nOptions:\n");
    printf("  -i           interactive - prompt before deleting\n");
    printf("  -v           verbose - display files that have been deleted\n");
    printf("  -? -h        display this help message\n");
    printf("  -V           show version number\n");
			    
    printf("\nExamples:\n\n");
    printf(" dndel -l 'myvax::*.*;'\n");
    printf(" dndel 'cluster\"patrick thecats\"::disk$users:[cats.pics]otto.jpg;'\n");
    printf("\n");
}
/*-------------------------------------------------------------------------*/
static int dap_setup_link(void)
{
    int sockfd;
    
    if ((sockfd=socket(AF_DECnet,SOCK_SEQPACKET,DNPROTO_NSP)) == -1) 
    {
	perror("socket");
	exit(-1);
    }

    
    if (setsockopt(sockfd,DNPROTO_NSP,SO_CONACCESS,&accessdata,
		   sizeof(accessdata)) < 0) 
    {
	perror("setsockopt");
	exit(-1);
    }
    
    sockaddr.sdn_family = AF_DECnet;
    sockaddr.sdn_flags	= 0x00;
    sockaddr.sdn_objnum	= 0x11;			/* FAL		*/
    sockaddr.sdn_objnamel	= 0x00;
    memcpy(sockaddr.sdn_add.a_addr, binadr->n_addr,6);

    if (connect(sockfd, (struct sockaddr *)&sockaddr, 
		sizeof(sockaddr)) < 0) 
    {
	perror("socket");
	exit(-1);
    }
    return sockfd;
}
/*-------------------------------------------------------------------------*/
static void dap_close_link(int sockfd)
{
    if (close(sockfd) < 0)
	perror("close");
}
/*-------------------------------------------------------------------------*/
static void	dap_exchg_config(int sockfd)
{
    int		er;
    unsigned char	buf[100];
    unsigned char	confmsg[17] = {
	0x01,	/* Message type 1 = Conf */
	0x00,	/* Flag Field		 */
	0x2E,0x04, /* BufSiz 		 */
	0x07,	/* OS type = VMS	 */
	0x03,   /* RMS-32		 */
	7,0,0,0,0, /* Version 'stuff'     */
	0xA2,	/* Sequential Org Support*/
	0x80,
	0xB0,   /* Message Blocking. OK  */
	0xFE,   /* Supp Extended Attrib  */
	0xA8,   /* Supp. Seq Access/Del  */
	0x7C	/* WildCrds/Ren/Seg Msgs */
    };
    
    if ((er=write(sockfd,confmsg,sizeof(confmsg))) < 0)
    {
	perror("Send DAP config msg");
	exit(-1);
    }			      
    if ((er=read(sockfd,buf,sizeof(buf))) < 0)
    {
	perror("Recv DAP config msg");
	exit(-1);
    }			      
    /* Assume that Peer Supports the minimal characteristis we
       setup here
    */
    if (buf[0] == 0x09)
    {
	printf("Local DAP configuration rejected by remote node\n");
	exit(-1);
    }
}
/*-------------------------------------------------------------------------*/
static	void	dap_directory_lookup(int sockfd, int interactive)
{
    unsigned char	part1[4] = {0x03,		/* Access TYPE	   */
				    0x00,		/* FLAGS	   */
				    0x06,		/* Directory Func  */
				    0x01 		/* I/O err is fatal*/
    }; 
    unsigned char	part2[4] = {0x00,		/* Default Acc GET */
				    0x00,		/* Share Ops GET   */
				    0x01,		/* Main attr only  */
				    0x00 		/* Password Length */
    }; 
    int	i,er;
    
    if (!interactive) part1[2] = 0x04; // $Erase

    for (i = 0; i < 4; i++) buf[i] = part1[i];
    buf[4] = strlen(dirname);
    for (i = 0; i < strlen(dirname); i ++) buf[i+5] = dirname[i];
    for (i = 0; i < 4; i++) buf[i+5+strlen(dirname)] = part2[i];
    
    if ((er=write(sockfd,buf,(strlen(dirname)+9))) < 0)
    {
	perror("Send DAP Directory Access");
	exit(-1);
    }
    buflen = bufptr = 0;
}

static	void	dap_delete_file(int sockfd, char *name)
{
    char buf[128];
    
    unsigned char	part1[]  = {0x03,		/* Access TYPE	   */
				    0x00,		/* FLAGS	   */
				    0x04,		/* $ERASE          */
				    0x01 		/* I/O err is fatal*/
    }; 
    unsigned char	part2[]  = {0x00,		/* Default Acc GET */
				    0x00,		/* Share Ops GET   */
				    0x00,		/* No attr         */
				    0x00 		/* Password Length */
    }; 
    int	i,er;
    int ptr = 0;
    
    for (i = 0; i < sizeof(part1); i++) buf[ptr++] = part1[i];
    buf[ptr++] = strlen(name);
    for (i = 0; i < strlen(name); i++) buf[ptr++] = name[i];
    for (i = 0; i < sizeof(part2); i++) buf[ptr++] = part2[i];
    
    if ((er=write(sockfd,buf,ptr)) < 0)
    {
	perror("Send DAP Delete");
	exit(-1);
    }
}
/*-------------------------------------------------------------------------*/
static int dap_get_delete_ack(int sockfd)
{
    char buf[128];
    int  ptr = 0;
    int er;
    
    if ((er=read(sockfd,buf,sizeof(buf))) < 0)
    {
	perror("Recv DAP ack");
	exit(-1);
    }
    if (buf[ptr++] == 0x09) // STATUS
    {
	printf("Error deleting file\n");
	return FALSE;
    }
    if (buf[ptr++] == 0x07) // ACCOMP
    {
	printf("got accomp\n");
    }
    return TRUE;
}
/*-------------------------------------------------------------------------*/
static void	dap_send_skip(int sockfd)
{	
    unsigned char	contmsg[3] = {0x05,0x00,0x02};
    int		er;
    
    if ( (er=write(sockfd,contmsg,3)) < 0)
    {
	perror("Send DAP skip message");
	exit(-1);
    }
    buflen = bufptr = 0;
}
/*-------------------------------------------------------------------------*/
static void	adjust_bufptr(int sockfd)
{
    if (buflen <= bufptr)
    {
	buflen=read(sockfd,buf,sizeof(buf));
	if (buflen < 0) 
	{
	    perror("DAP Read Directory Entry");
	    exit(-1);
	}
	bufptr=0;
    }
}
/*-------------------------------------------------------------------------*/
static int	dap_get_dir_entry(int sockfd, char *name)
#define	BC	buf[bufptr]
#define MC	menu[0] 
{
    int	        retval,len,i,auxptr;
    char	menu[6];
    static char volname[255], dirname[255], filname[255];

    adjust_bufptr(sockfd);
    while ((BC != 0x06) && (BC != 0x07) )
    {
	adjust_bufptr(sockfd);
	switch (BC) {
	case 0x09 : if ((buf[bufptr+2] == 0x30) && 
			(buf[bufptr+3] == 0x40) )
	{
	    printf("%s %s\n",name,
		   "(Locked by another user)");
	    dap_send_skip(sockfd);
	    break;
	}
	else return -BC;

	case 0x0F : bufptr += 3; /* NAM block */
	    if (BC == 0x08)		/* VOL Name */
	    {
		bufptr += 1;
		memcpy(volname,&buf[bufptr+1],BC);
		volname[BC] = '\0';
		bufptr += (BC + 1);
		break;
	    }
	    if (BC == 0x04)		/* DIR Name */
	    {
		bufptr += 1;
		memcpy(dirname,&buf[bufptr+1],BC);
		dirname[BC] = '\0';
		bufptr += (BC + 1);
		break;
	    }
	    if (BC == 0x02)		/* FileName */
	    {
		memset(filname,0,80);
		bufptr += 1;
		memcpy(filname,&buf[bufptr+1],BC);
		filname[BC] = '\0';
		bufptr += (BC + 1);
		sprintf(name, "%s%s%s", volname, dirname, filname);
		break;
	    }
	    break;
	case 0x02 :
	    adjust_bufptr(sockfd);
	    bufptr += 2;
	    len = BC+1;
	    auxptr = bufptr;
	    i=0; bufptr += 1;
	    while (BC & 0x80)
	    {
		menu[i++] = BC;
		bufptr += 1;
	    }
	    bufptr += 1;
	    if (MC & 1)	/* DataType */
	    {
		bufptr += BC+1;
	    }
	    if (MC & 2)	/* ORG		*/
	    {
		bufptr += 1;
	    }
	    if (MC & 4)	/* RFM		*/
	    {
		bufptr += 1;
	    }
	    if (MC & 8)	/* RAT		*/
	    {
		while (BC & 0x80)
		    bufptr += 1;
		bufptr += 1;
	    }
	    if (MC & 16)	/* BLS		*/
	    {
		bufptr += 2;
	    }
	    if (MC & 32)	/* MRS		*/
	    {
		bufptr += 2;
	    }
	    bufptr = auxptr + len;
	    break;

	case 0x0D :
	    adjust_bufptr(sockfd);
	    bufptr += 2;
	    len = BC+1;
	    bufptr += len;
	    break;
	default: 
            bufptr += 1;
            adjust_bufptr(sockfd);
	}
	adjust_bufptr(sockfd);
    }

    if (BC == 0x06)
    {
	bufptr += 1;
	if (BC == 0x02)
	    bufptr += 2;
	else    bufptr += 1;
	return 0x01;
    }
    if (BC == 0x07)
    {
        //printf("PJC: ACCOMP code is 0x%x\n", buf[bufptr+2]);
        return 0x00; //ACCOMP
    }
//    printf("PJC: Returning error 0x%x 0x%x (bufptr=%d, buflen=%d)\n", BC, buf[bufptr+1], bufptr, buflen);
    return -BC;
}
/*-------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
    int	    opt,retval;
    char    name[256];
    int     interactive = 0;
    int     verbose = 0;
    int     two_links = 0;
    int     dir_sock, del_sock;

    /*	Zero Structure	AccessData			*/
    memset(&accessdata,0,sizeof(accessdata));

    if (argc < 2)
    {
	usage();
	exit(0);
    }


    
/* Get command-line options */
    opterr = 0;
    optind = 0;
    while ((opt=getopt(argc,argv,"?hvVi")) != EOF)
    {
	switch(opt)
	{
	case 'h': 
	case '?':
	    usage();
	    exit(1);

	case 'i':
	    interactive++;
	    two_links++;
	    break;

	case 'v':
	    verbose++;
	    two_links++;
	    break;

	case 'V':
	    printf("\ndndel from dntools version %s\n\n", VERSION);
	    exit(1);
	    break;
	}
    }

    /* Parse the filename into its parts */
    if (!parse(argv[optind]))
    {
	fprintf(stderr, "%s\n", lasterror);
	exit(2);
    }

    /* We open one link to get the file names and (if interactive or verbose)
       another to do the actual deletion.
       This is not a waste of resources - it's what VMS does !
     */
    dir_sock = dap_setup_link();        /* Setup link for dir	*/
    dap_exchg_config(dir_sock);


    /* Open the second link to actually delete the file(s) */
    if (two_links)
    {
	del_sock = dap_setup_link();        /* Setup link for del	*/
	dap_exchg_config(del_sock);
    }
    
    /* If non-interactive then the next command just deletes the files */
    dap_directory_lookup(dir_sock, interactive || verbose); 
    
    /* Loop through the files we find */
    while ((retval=dap_get_dir_entry(dir_sock, name)) > 0)
    {
	if (retval == 0x01)
	{
	    if (interactive)
	    {
		char response[255];
		
		printf("Delete %s ? ", name);
		fgets(response, sizeof(response), stdin);
		if (tolower(response[0]) == 'y')
		{
		    dap_delete_file(del_sock, name);
		    if (dap_get_delete_ack(del_sock))
			if (verbose)
			    printf("Deleted %s\n", name);
		}
	    }		
	    else if (verbose)
	    {
		dap_delete_file(del_sock, name);
		if (dap_get_delete_ack(del_sock))
		    printf("Deleted %s\n", name);
	    }
	}
    }
    if (retval == -9) printf("Error deleting %s\n",dirname);
    dap_close_link(dir_sock);
    if (two_links)
	dap_close_link(del_sock);
    return 0;
}
/*-------------------------------------------------------------------------*/
/* Parse the filename into its component parts */
static int parse(char *fname)
{
    enum  {NODE, USER, PASSWORD, ACCOUNT, NAME, FINISHED} state;
    int   n0=0;
    int   n1=0;
    char  node[MAX_NODE+1];
    char *nodenum;
    char *local_user;
    char  lastchar;
    
    state = NODE; /* Node is mandatory */

    while (state != FINISHED)
    {
	switch (state)
	{
	case NODE:
	    if (fname[n0] != ':' && fname[n0] != '\"' && fname[n0] != '\'')
	    {
		if (n1 >= MAX_NODE || 
		    fname[n0] == ' ' || fname[n0] == '\n')
		{
		    lasterror = "File name parse error";
		    return FALSE;
		}
		node[n1++] = fname[n0++];
	    }
	    else
	    {
		node[n1] = '\0';
		n1 = 0;
		if (fname[n0] == '\"')
		{
		    n0++;
		    state = USER;
		}
		else
		{
		    n0 += 2;
		    state = NAME;
		}
	    }
	    break;

	case USER:
	    if (fname[n0] != ' ' && fname[n0] != '\"' && fname[n0] != '\'')
	    {
		if (n1 >= MAX_USER)
		{
		    lasterror = "File name parse error";
		    return FALSE;
		}
		accessdata.acc_user[n1++] = fname[n0++];
	    }
	    else
	    {
		accessdata.acc_user[n1] = '\0';
		n1 = 0;
		if (fname[n0] == ' ')
		{
		    state = PASSWORD;
		    n0++;
		}
		else /* Must be a quote */
		{
		    state = NAME;
		    /* Check for :: */
		    n0 += 3;
		}
	    }
	    break;

	case PASSWORD:
	    if (fname[n0] != ' ' && fname[n0] != '\"' && fname[n0] != '\'')
	    {
		if (n1 >= MAX_PASSWORD)
		{
		    lasterror = "File name parse error";
		    return FALSE;
		}
		accessdata.acc_pass[n1++] = fname[n0++];
	    }
	    else
	    {
		accessdata.acc_pass[n1] = '\0';
		n1 = 0;
		if (fname[n0] == ' ')
		{
		    n0++;
		    state = ACCOUNT;
		}
		else /* Must be a quote */
		{
		    state = NAME;
		    /* Check for :: */
		    n0 += 3;
		}
	    }
	    break;

	case ACCOUNT:
	    if (fname[n0] != '\'' && fname[n0] != '\"')
	    {
		if (n1 >= MAX_ACCOUNT)
		{
		    lasterror = "File name parse error";
		    return FALSE;
		}
		accessdata.acc_acc[n1++] = fname[n0++];
	    }
	    else
	    {
		accessdata.acc_acc[n1] = '\0';;
		state = NAME;
		n1 = 0;
		/* Check for :: */
		n0 += 3;
	    }
	    break;

	case NAME:
	    strcpy(dirname, fname+n0);
	    state = FINISHED;
	    break;

	case FINISHED: // To keep the compiler happy
	    break;
	} /* switch */
    }

/* tail end validation */
    binadr = getnodebyname(node);
    if (!binadr)
    {
	lasterror = "Unknown or invalid node name ";
	return FALSE;
    }

    lastchar = dirname[strlen(dirname)-1];
    if (lastchar == ']' || lastchar == ':' || dirname[0] == '\0')
	fprintf(stderr, "No filename given\n");

    // Try very hard to get the local username
    local_user = getlogin();
    if (!local_user || local_user == (char *)0xffffffff)
      local_user = getenv("LOGNAME");
    if (!local_user) local_user = getenv("USER");
    if (local_user)
    {
        int i;

	strcpy((char *)accessdata.acc_acc, local_user);
	accessdata.acc_accl = strlen((char *)accessdata.acc_acc);
        for (i=0; i<accessdata.acc_accl; i++)
            accessdata.acc_acc[i] = toupper(accessdata.acc_acc[i]);
    }
    else
        accessdata.acc_acc[0] = '\0';


    /* Complete the accessdata structure */
    accessdata.acc_userl = strlen(accessdata.acc_user);
    accessdata.acc_passl = strlen(accessdata.acc_pass);
    accessdata.acc_accl  = strlen(accessdata.acc_acc);
    return TRUE;
}

