/******************************************************************************
    (c) 1998 P.J. Caulfield               patrick@pandh.demon.co.uk
                                          pcaulfield@cix.co.uk
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 ******************************************************************************

 This file is based heavily on Eduardo's original dncopy program supplied with
 the kernel patch for Linux 1.0

 ******************************************************************************
*/
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include <netdnet/dn.h>
#include <netdnet/dnetdb.h>
#include "dnetfile.h"


#define	BC	buf[bufptr]
#define MC	menu[0]  
#define MC1	menu[1]
#define MC2	menu[2]

/*-------------------------------------------------------------------------*/
int dnetfile::dap_setup_link()
{
    if (verbose > 2) printf("in dap_setup_link()\n");

    if ((sockfd=socket(AF_DECnet,SOCK_SEQPACKET,DNPROTO_NSP)) == -1) 
    {
        sprintf(errstring, "socket failed: %s", strerror(errno));
	lasterror = errstring;
	return -1;
    }
    
    if (setsockopt(sockfd, DNPROTO_NSP, SO_CONACCESS, &accessdata,
		   sizeof(accessdata)) < 0) 
    {
        sprintf(errstring, "setsockopt failed: %s", strerror(errno));
	lasterror = errstring;
	return -1;
    }
    
    sockaddr.sdn_family = AF_DECnet;
    sockaddr.sdn_flags	= 0x00;
    sockaddr.sdn_objnum	= 0x11;			/* FAL		*/
    sockaddr.sdn_objnamel	= 0x00;
    memcpy(sockaddr.sdn_add.a_addr, binadr->n_addr,6);
    
    if (connect(sockfd, (struct sockaddr *)&sockaddr, 
		sizeof(sockaddr)) < 0) 
    {
        sprintf(errstring, "connect failed: %s", strerror(errno));
	lasterror = errstring;
	return -1;
    }

    int flags = fcntl(sockfd, F_GETFL, 0);
    fcntl(sockfd, F_SETFL, flags & ~O_NONBLOCK);

    bufptr = buflen = 0;
    return 0;
}

/*-------------------------------------------------------------------------*/
void dnetfile::dap_close_link()
{
    if (verbose > 2) printf("in dap_close_link()\n");
    ::close(sockfd);
}

/*-------------------------------------------------------------------------*/
int dnetfile::dap_exchg_config()
{
    if (verbose > 2) printf("in dap_exchg_config()\n");

    int		er;
    int         bptr=0;
    unsigned char	confmsg[] = {
	0x01,	/* Message type 1 = Conf */
	0x00,	/* Flag Field		 */
	MAX_READ_SIZE%256,
	MAX_READ_SIZE/256,   /* BufSiz 	 */
	0x07,	/* OS type = VMS	 */
	0x03,   /* RMS-32		 */
	7,0,0,0,0, /* Version 'stuff'     */
	0xA2,	/* Sequential Org Support*/
	0x80,
	0xB0,   /* Message Blocking. OK  */
	0xFE,   /* Supp Extended Attrib  */
	0xA8,   /* Supp. Seq Access/Del  */
	0x7C	/* WildCrds/Ren/Seg Msgs */
    };
    if ((er=::write(sockfd,confmsg,sizeof(confmsg))) < 0)
    {
        sprintf(errstring, "Send DAP config msg failed: %s", strerror(errno));
	lasterror = errstring;
	return -1;
    }

    if (!adjust_bufptr()) return -1;
    
    /* Assume that Peer Supports the minimal characteristics we
       setup here
    */
    if (buf[0] == 0x09)
    {
	lasterror = "Local DAP configuration rejected by remote node";
	return -1;
    }
    bptr = 2;
    if (buf[1] & 0x02) bptr++; // Message has a length byte

    // Get the remote buffer size and use that as the maximum allowable
    // block/record size. Subtract three from the number we receive back
    // to allow for the header to a PUT command
    remote_bufsize = buf[bptr+1]*256 + buf[bptr] - 3;
    if (verbose > 1) printf("remote buffer size = %d\n", remote_bufsize);
    return 0;
}
/*-------------------------------------------------------------------------*/

// This is where we do most of the reading from the remote end. VMS blocks
// several requests into one buffer so we move through the buffer here and
// read another block then that one has been exhausted.
bool dnetfile::adjust_bufptr(void)
{
    if (buflen <= bufptr || BC == '\0')
    {
	do
        {
 	    buflen=::read(sockfd,buf,MAX_READ_SIZE);
        } while (buflen == 0);
	if (buflen < 0) 
	{
	    sprintf(errstring, "DAP read error: %s", strerror(errno));
	    lasterror = errstring;
	    return FALSE;
	}
	if (verbose >2) printf("adjust_bufptr: read %d bytes\n", buflen);
	bufptr=0;
    }
    return TRUE;
}
/*-------------------------------------------------------------------------*/
int dnetfile::dap_get_status(void)
{
    if (!adjust_bufptr()) return -1;    

    if (BC == 0x09)
    {
        return dap_check_status(0);
    }
    fprintf(stderr, "dap_get_status: wrong code type : %x\n", BC);
    lasterror = "Unknown message type received";
    return -1; 
}
/*-------------------------------------------------------------------------*/
/* Check that we have at least 'needed' bytes in the buffer
 * This only seems to be called for STREAM files where VMS sends the file
 * as a "stream" (surprise, surprise) rather than a collection of
 * records 
 */
void dnetfile::check_length(int needed)
{
    if (bufptr+needed > buflen)
    {
	int left = buflen - bufptr; /* what's left unread */

	/* Move what we have to the start of the buffer */
	memmove(buf, buf+bufptr, left);

	if (verbose > 2) printf("check_length(): reading up to %d bytes to fill record. bufptr=%d, buflen=%d\n",
				needed - left, bufptr, buflen);

	/* read enough to satisfy what's needed */
	buflen = ::read(sockfd, buf+left, MAX_READ_SIZE);
	if (buflen < 0) 
	{
	    perror("DAP Read Error");
	    return;
	}
	buflen += left;
	bufptr=0;
    }
}
/*-------------------------------------------------------------------------*/
int dnetfile::dap_send_access()
{
    int ptr = 0;
    if (verbose > 2) printf("in dap_send_access(%s)\n", filname);

    unsigned	char	attmsg[]  = {0x02,	/* Att. Type	*/
				     0x02,	/* Len present  */
				     0x01,
				     0x00 	/* No att field */
    };

    unsigned	char	acc1[]    = {0x03,   /* Acc Message  */
				     0x00,
				     0x01,   /* Open Function*/
				     0x01    /* no  fatal err*/
    };
    unsigned	char	acc2[]    = {0x02,   /* FB$GET access */
				     0x02,   /* FB$GET sharing */
				     0x31,   /* Get date/time & prot */
				     0x0};   /* No file password */
    unsigned    int	i;
    int                 er;

    if (transfer_mode == MODE_BLOCK)
    {
	acc2[0] = 0x22; // Block access
	acc2[1] = 0x40; // No sharing
    }

    if (writing)
    {
	acc1[2] = 0x02; // $CREATE
	acc2[0] = 0x01; // $PUT
	acc2[1] = 0x40; // No sharing
    }
    else // We build our own attributes message when writing.
    {
 	for (i=0; i < sizeof(attmsg); i++)  buf[ptr++] = attmsg[i];
    }
    
    for (i=0; i < sizeof(acc1); i++)    buf[ptr++] = acc1[i];
    buf[ptr++] = strlen(filname);
    for (i=0; i < strlen(filname); i++) buf[ptr++] = filname[i];
    for (i=0; i < sizeof(acc2); i++)    buf[ptr++] = acc2[i];

    if ( (er=::write(sockfd,buf,ptr)) < 0)
    {
        sprintf(errstring, "Send DAP access msg failed: %s", strerror(errno));
	lasterror = errstring;
	return -1;
    }
    buflen = bufptr = 0;
    return 0;
}

/*-------------------------------------------------------------------------*/

// Receive the attributes of a file for download or a newly created file
int dnetfile::dap_get_file_entry(int *rfm, int *rat)
{
    if (verbose > 2) printf("in dap_get_file_entry()\n");
    int	retval,len,i,auxptr;
    char	menu[6];
    
    if (!adjust_bufptr()) return -1;    
    while ( (BC != 0x06) && (BC != 0x07) && (BC != 0x09) ) 
    {
        if (!adjust_bufptr()) return -1;

	switch(BC)
	{
	case 0x0F :  /* NAM block */
	    bufptr++; if (BC & 2) bufptr++;
	    bufptr++;
	    if (BC == 0x08)		/* VOL Name */
	    {
		bufptr += 1;
		memcpy(volname,&buf[bufptr+1],BC);
		bufptr += (BC + 1);
		break;
	    }
	    if (BC == 0x04)		/* DIR Name */
	    {
		bufptr += 1;
		memcpy(dirname,&buf[bufptr+1],BC);
		bufptr += (BC + 1);
		break;
	    }
	    if (BC == 0x02)		/* FileName */
	    {
		memset(filname,0,80);
		bufptr += 1;
		memcpy(filname,&buf[bufptr+1],BC);
		bufptr += (BC + 1);
		if (verbose > 3)
		    printf("Got file name %s\n", filname);
		break;
	    }
	case 0x02 :  /* File attributes */
	    bufptr++; if (BC & 2) bufptr++;
	    len = BC + 1;
	    auxptr = bufptr;
	    i=0; bufptr += 1;

	    while (BC & 0x80) 
	    {
		menu[i++]= BC;
		bufptr += 1;
	    }
	    menu[i++]= BC;
	    bufptr += 1;
	    if (MC & 1) 		/* Datatype	*/
	    {
		bufptr += BC + 1;
	    }
	    if (MC & 2)			/* ORG		*/
	    {
		bufptr++;
	    }
	    if (MC & 4)			/* RFM	*/
	    {
		*rfm = BC;
		bufptr++;
	    }
	    if (MC & 8)			/* RAT		*/
	    {
	        *rat = BC;              // Only need the first byte's worth
		while (BC & 0x80) bufptr++;
		bufptr++;
	    }
            if (MC & 16)                 /* BLS          */
            {
		bufptr += 2;
	    }
            if (MC & 32)                 /* MRS          */
            {
                bufptr += 2;
            }
            if (MC & 64)                 /* ALQ          */
            {
                bufptr += BC+1; 
            }
	    if (MC1 & 1)                /* BKS          */
	    {
		bufptr++;
	    }
            if (MC1 & 2)                /* FSZ          */
            {
                file_fsz = BC;
                bufptr++;
            }
	    if (MC1 & 4)               /* MRN           */
	    {
		bufptr += BC+1;
	    }
            if (MC1 & 8)               /* RUNSYS         */
	    {
		bufptr += BC+1;
	    }
            if (MC1 & 16)              /* DEQ        */
	    {
		bufptr += 2;
	    }
	    if (MC1 & 32)              /* FOP        */
	    {
                while (BC & 0x80) bufptr++;
                bufptr++;
	    }
            if (MC1 & 64)              /* BSZ        */
	    {
		bufptr++;
	    }
            if (MC2 & 1)              /* DEV        */
	    {
                while (BC & 0x80) bufptr++;
                bufptr++;
	    }
            if (MC2 & 2)              /* SDC        */
	    {
                while (BC & 0x80) bufptr++;
                bufptr++;
	    }
            if (MC2 & 4)              /* LRL        */
	    {
                bufptr += 2;
	    }
            if (MC2 & 8)               /* HBK         */
	    {
		bufptr += BC+1;
	    }
            if (MC2 & 16)              /* EBK         */
	    {
		bufptr += BC+1;
	    }
            if (MC2 & 32)              /* FFB        */
	    {
                bufptr += 2;
	    }
            if (MC2 & 64)               /* SBN         */
	    {
		bufptr += BC+1;
	    }
	    break;

        case 0x0D: /* Date/time stamps */
	  bufptr++; if (BC & 2) bufptr++;
	  len = BC;
	  bufptr += len + 1;
	  break;

        case 0x0E :   // File protection
	  bufptr++; if (BC & 2) bufptr++;
	  bufptr += 2;
	  bufptr += (BC + 1);
	  prot = 0;
	  for (i=0; i < 4; i++)
	  {
	    if (i>0) // SYSTEM protection not supported on Unix
	    {
	      prot = prot<<3;
	      if ((BC & 0x01)==0) prot=prot|4;
	      if ((BC & 0x02)==0) prot=prot|2;
	      // neither is Delete permission
	      if ((BC & 0x04)==0) prot=prot|1;
	    }
	    bufptr += 1;
	  }
	  break;

	case 0x07:
	  continue; // Reached end of wildcard list

	case 0x06:
	  break;

	default:
	  bufptr += 1;
          if (!adjust_bufptr()) return -1;
	}
    }

    if (BC == 0x06) // ACK
    {
	bufptr += 1;
	if (BC == 0x02)
	    bufptr += 2;
	else
	    bufptr += 1;
	return 0;
    }
    
    if (BC == 0x09) // STATUS
    {
        return dap_check_status(0);
    }
    if (BC == 0x07)
    {
        lasterror = "End of wildcard list";
        return -2;
    }

    sprintf(errstring, "unrecognised tag in get_file_entry: %d", BC);
    lasterror=errstring;
	
    return -1;
}
/*-------------------------------------------------------------------------*/
int dnetfile::dap_send_connect()
{
    if (verbose > 2) printf("in dap_send_connect()\n");
    unsigned char	conmsg[4]  = {0x04,0x00,0x02,0x00};
    int	er;
    
    if ((er=::write(sockfd,conmsg,4)) < 0)
    {
        sprintf(errstring, "Send DAP connect msg failed: %s", strerror(errno));
	lasterror = errstring;
	return -1;
    }
    
    if (!adjust_bufptr()) return -1;
    if (BC == 0x09)
    {
        return dap_check_status(0);
    }
    if (BC != 0x06)
    {
	sprintf(errstring,"DAP send connect did not receive ACK (got %x)", BC);
	lasterror = errstring;
	return -1;
    }
    bufptr += 2;
    return 0;
}
/*-------------------------------------------------------------------------*/
// Sends a CONTROL message with $GET or $PUT as the action
// Also enables block mode for reading files with -mblock requested.
// Note we don't really write files in block mode only read them.
// Writes just create fixed-length record binary files.
int dnetfile::dap_send_get_or_put()
{
    if (verbose > 2) printf("in dap_send_get_or_put()\n");
    unsigned char	conmsg[]  = {0x04, /* CONTROL message */
                                     0x02, /* length present */
                                     0x03, /* Length */
                                     0x01, /* Function 1=$GET */
                                     0x01, /* Menu: 1=RAC */
                                     0x03  /* RAC: 3=SEQ transfer */
                                    };
    int	er;

    if (transfer_mode == MODE_BLOCK && !writing)
    {
	conmsg[5] = 0x05; // Get blocks rather than records
    }

    if (writing)
    {
	conmsg[3] = 0x04; // $PUT
    }
    
    if ((er=::write(sockfd,conmsg,sizeof(conmsg))) < 0)
    {
        sprintf(errstring, "Send DAP get msg failed: %s", strerror(errno));
	lasterror = errstring;
	return -1;
    }
    return 0;
}
/*-------------------------------------------------------------------------*/

// Send Access Complete.
int dnetfile::dap_send_accomp()
{
    if (verbose > 2) printf("in dap_send_accomp()\n");
    unsigned char	acomsg2[] = {0x07,0x00,0x01,0x00}; // $CLOSE
    int	er;

    if ((er=::write(sockfd,acomsg2,sizeof(acomsg2))) < 0)
    {
        sprintf(errstring, "Send DAP access complete msg failed: %s", strerror(errno));
	lasterror = errstring;
	return -1;
    }

    // ACCOMP gets a reply, either a STATUS message or just an ACK
    if (!adjust_bufptr()) return -1;
    if (BC == 0x09) return dap_check_status(0);
    if (BC == 0x07) // Got ACK to ACCOMP
    {
        bufptr += 3;
        if (buf[bufptr-1] == 0x02)
            wildcard = false; // reached end of wildcard list
    }
    return 0;
}
/*-------------------------------------------------------------------------*/
int dnetfile::dap_get_record(char *rec, int reclen)
{
    if (verbose > 2) printf("in dap_get_record() bufptr=%d, buflen=%d\n", bufptr, buflen);
    int	len;

    if (!adjust_bufptr()) return -1;
    if (BC == 0x09) 
    {
	int status = dap_check_status(0);
	if (status == 047) 
	{
	    ateof = TRUE;
	    return -1;
	};
	if (status == -1) return -1;

	if (!adjust_bufptr()) return -1;
    }
    if (BC != 0x08) // Not data or status!!
    {
        sprintf(errstring, "Unknown block type (%d) received", BC);
	lasterror = errstring;
	return -1;
    }

    bufptr++;

// This next line is not redundant !! 
    if (bufptr >= buflen && !adjust_bufptr()) return -1;
 
    bool use_buflen = false; 
    if (! (BC & 2)) // No length field - use the whole block
    {
	if (BC & 1) bufptr++; // STREAMID present
	len = buflen - bufptr -2;
	use_buflen = true;
    }
    else
    {    
       bufptr++;
       if (!adjust_bufptr()) return -1;
       len = BC-1;
    }

    if (verbose > 2) printf("dap_get_record: len = %d\n", len); 
    if (len == -1) 
    {
        return 0; // Empty record??
    }
    if (bufptr >= buflen && !adjust_bufptr()) return -1;

    if (!use_buflen) check_length(len+2); 
    bufptr++;
    bufptr++; 
    
    memset(rec,0,len+1);
    memcpy(rec, &buf[bufptr], len);
    if (use_buflen) 
        bufptr = buflen;
    else
        bufptr += len;

    if (verbose > 1) printf("got record: %d bytes\n",len);

    return len;
}

// Send a record to VMS
int dnetfile::dap_put_record(char *rec, int reclen)
{
    if (verbose > 2) printf("in dap_put_record(%d bytes)\n", reclen);

    char writebuf[reclen+3];
    int  status;

    writebuf[0] = 0x08; // Data message
    writebuf[1] = 0x00; // Length not present
    writebuf[2] = 0x00; // Attr
    memcpy(&writebuf[3], rec, reclen);

    status = ::write(sockfd, writebuf, reclen+3);
    if (!status) lasterror = strerror(errno);
    
    if (verbose > 2) printf("Written %d bytes\n", status);

    int oob_status = dap_check_oob(); 
    if (oob_status) return oob_status;
    return status;
}

// Send the attributes for a newly created file
int dnetfile::dap_send_attributes()
{
    if (verbose > 2) printf("in dap_send_attributes()\n");
    unsigned	char	attmsg[] = {0x02,	/* Att. Type	*/
				    0x02,	/* Len present  */
				    0x0c,	/* Length       */
				    0xee,0x20, 	/* Attributes sent */
				    0x00,       /* ORG=SEQ      */
				    RFM_STMLF,
				    RAT_CR,
				    0, 0,       /* MRS set later */
			            0x01, 0x02, /* ALQ */	
				    0x80, 0x80, 0x10, /* FOP=CIF, SQO */
/* This is an ALLOC message: */
	0x0b, 0x02, 0x0d, 0xef, 0x03, 0x00,0x00,0x00,0x00,0x00,0x01,0x02,
	0x00, 0x00,0x00,0x00

    };
    int	                i,er;

// No Record attributes on BLOCK sends
    if (transfer_mode == MODE_BLOCK)
    {
	attmsg[7] = RAT_NONE;
    }

    attmsg[8] = user_bufsize%256;
    attmsg[9] = user_bufsize/256;

// Set the attributes of the uploaded file
    if (user_rat != RAT_DEFAULT)
	attmsg[7] = user_rat;

    if (user_rfm != RFM_DEFAULT)
	attmsg[6] = user_rfm;

    if ( (er=::write(sockfd,attmsg,sizeof(attmsg))) < 0)
    {
        sprintf(errstring, "Send DAP attributes msg failed: %s", strerror(errno));
	lasterror = errstring;
	return -1;
    }
    return 0;
}

int dnetfile::dap_send_name()
{
    if (verbose > 2) printf("in dap_send_name() %s\n", name);
    unsigned	char	nammsg[] = {0x0E,	/* Name Type	*/
				    0x00,	/* Len not present  */
				    0x00, 	/* No att field */
				    0x01, 	/* Full file spec */
				    0x00, 	/* space for name length*/
    };
    int	                i,er;

    nammsg[4] = strlen(name);
    memcpy(buf, nammsg, sizeof(nammsg));
    memcpy(buf+sizeof(nammsg), name, strlen(name));
    
    if ( (er=::write(sockfd,buf,sizeof(nammsg)+strlen(name))) < 0)
    {
        sprintf(errstring, "Send DAP name msg failed: %s", strerror(errno));
	lasterror = errstring;
	return -1;
    }
    return 0;
}

// Look for any out-of-band responses and deal with them as best we can
int dnetfile::dap_check_oob()
{
    int len;
    int flags;

    flags = fcntl(sockfd, F_GETFL, 0);
    fcntl(sockfd, F_SETFL, flags | O_NONBLOCK);
    len = ::read(sockfd,buf+bufptr,MAX_READ_SIZE);
    if (len > 0)
    {
        if (verbose > 2) printf("got %d oob bytes, bufptr is %d, BC=%x\n",
				len, bufptr, BC);
	buflen += len;
	fcntl(sockfd, F_SETFL, flags);
	if (BC == 0x09) return dap_check_status(0);
        fprintf(stderr, "unknown response %x received\n", BC);
	return -1;
    }
    fcntl(sockfd, F_SETFL, flags);
    return 0;
}

// Skip to the next file. (used for skipping directories)
int dnetfile::dap_send_skip()
{
    unsigned char  contmsg[] = {0x05,0x00,0x02};
    int	           er;
    
    if ( (er=::write(sockfd,contmsg, sizeof(contmsg))) < 0)
    {
        sprintf(errstring, "Send DAP skip message: %s", strerror(errno));
	lasterror = errstring;
	return -1;
    }
    return 0;
}

// Check a STATUS message. code 0225 is Success and 047 is EOF so we return a
// correct status code for those cases.
int dnetfile::dap_check_status(int status)
{
    int code;

    bufptr++;  // move to flags
    if (BC & 2) bufptr++; // skip length
    bufptr++;  // skip flags

    code = BC;

    bufptr += 4; // TODO: This may not always be right: check the spec

    if (code == 0225) return status; // Success
    if (code == 047)  return code;   // EOF
    lasterror = dap_error(code);
    return -1;
}

// Convert a DAP error code into a message
char *dnetfile::dap_error(int code)
{
    switch (code)
    {
    case 01: return "operation aborted.";
    case 02: return "F11-ACP could not access file.";
    case 03: return "\"FILE\" activity precludes operation.";
    case 04: return "bad area ID.";
    case 05: return "alignment options error.";
    case 06: return "allocation quantity too large or 0 value.";
    case 07: return "not ANSI \"D\" format.";
    case 010: return "allocation options error.";
    case 011: return "invalid (i.e., synch) operation at AST level.";
    case 012: return "attribute read error.";
    case 013: return "attribute write error.";
    case 014: return "bucket size too large.";
    case 015: return "bucket size too large.";
    case 016: return "\"BLN\" length error.";
    case 017: return "beginning of file detected.";
    case 020: return "private pool address.";
    case 021: return "private pool size.";
    case 022: return "internal RMS error condition detected.";
    case 023: return "cannot connect RAB.";
    case 024: return "$UPDATE changed a key without having attribute of XB$CHG set.";
    case 025: return "bucket format check-byte failure.";
    case 026: return "RSTS/E close function failed.";
    case 027: return "invalid or unsupported \"COD\" field.";
    case 030: return "F11-ACP could not create file (STV=sys err code).";
    case 031: return "no current  record  (operation  not  preceded  by GET/FIND).";
    case 032: return "F11-ACP deaccess error during \"CLOSE\".";
    case 033: return "data \"AREA\" number invalid.";
    case 034: return "RFA-Accessed record was deleted.";
    case 035: return "bad device, or inappropriate device type.";
    case 036: return "error in directory name.";
    case 037: return "dynamic memory exhausted.";
    case 040: return "directory not found.";
    case 041: return "device not ready.";
    case 042: return "device has positioning error.";
    case 043: return "\"DTP\" field invalid.";
    case 044: return "duplicate key detected, XB$DUP not set.";
    case 045: return "RSX-F11ACP enter function failed.";
    case 046: return "operation not selected in \"ORG$\" macro.";
    case 047: return "end-of-file.";
    case 050: return "expanded string area too short.";
    case 051: return "file expiration date not yet reached.";
    case 052: return "file extend failure.";
    case 053: return "not a valid FAB (\"BID\" NOT = FB$BID).";
    case 054: return "illegal FAC for REC-OP,0, or FB$PUT not  set  for \"CREATE\".";
    case 055: return "file already exists.";
    case 056: return "invalid file I.D.";
    case 057: return "invalid flag-bits combination.";
    case 060: return "file is locked by other user.";
    case 061: return "RSX-F11ACP \"FIND\" function failed.";
    case 062: return "file not found.";
    case 063: return "error in file name.";
    case 064: return "invalid file options.";
    case 065: return "DEVICE/FILE full.";
    case 066: return "index \"AREA\" number invalid.";
    case 067: return "invalid IFI value or unopened file.";
    case 070: return "maximum NUM(254) areas/key XABS exceeded.";
    case 071: return "$INIT macro never issued.";
    case 072: return "operation illegal or invalid for file organization.";
    case 073: return "illegal record encountered (with sequential files only).";
    case 074: return "invalid ISI value, on unconnected RAB.";
    case 075: return "bad KEY buffer address (KBF=0).";
    case 076: return "invalid KEY field (KEY=0/neg).";
    case 077: return "invalid key-of-reference ($GET/$FIND).";
    case 0100: return "KEY size too large.";
    case 0101: return "lowest-level-index \"AREA\" number invalid.";
    case 0102: return "not ANSI labeled tape.";
    case 0103: return "logical channel busy.";
    case 0104: return "logical channel number too large.";
    case 0105: return "logical extend error, prior extend still valid.";
    case 0106: return "\"LOC\" field invalid.";
    case 0107: return "buffer mapping error.";
    case 0110: return "F11-ACP could not mark file for deletion.";
    case 0111: return "MRN value=neg or relative key>MRN.";
    case 0112: return "MRS value=0 for fixed length records.  Also 0 for relative files.";
    case 0113: return "\"NAM\"  block  address  invalid  (NAM=0,  or   not accessible).";
    case 0114: return "not positioned to EOF (sequential files only).";
    case 0115: return "cannot allocate internal index descriptor.";
    case 0116: return "indexed file; no primary key defined.";
    case 0117: return "RSTS/E open function failed.";
    case 0120: return "XAB'S not in correct order.";
    case 0121: return "invalid file organization value.";
    case 0122: return "error in file's prologue (reconstruct file).";
    case 0123: return "\"POS\" field invalid (POS>MRS,STV=XAB indicator).";
    case 0124: return "bad file date field retrieved.";
    case 0125: return "privilege violation (OS denies access).";
    case 0126: return "not a valid RAB (\"BID\" NOT=RB$BID).";
    case 0127: return "illegal RAC value.";
    case 0130: return "illegal record attributes.";
    case 0131: return "invalid record  buffer  address  (\"ODD\",  or  not word-aligned if BLK-IO).";
    case 0132: return "file read error.";
    case 0133: return "record already exists.";
    case 0134: return "bad RFA value (RFA=0).";
    case 0135: return "invalid record format.";
    case 0136: return "target bucket locked by another stream.";
    case 0137: return "RSX-F11 ACP remove function failed.";
    case 0140: return "record not found.";
    case 0141: return "record not locked.";
    case 0142: return "invalid record options.";
    case 0143: return "error while reading prologue.";
    case 0144: return "invalid RRV record encountered.";
    case 0145: return "RAB stream currently active.";
    case 0146: return "bad record size (RSZ>MRS,  or  NOT=MRS  if  fixed length records).";
    case 0147: return "record too big for user's buffer.";
    case 0150: return "primary  key  out  of  sequence  (RAC=RB$SEQ  for $PUT).";
    case 0151: return "\"SHR\"  field  invalid  for  file  (cannot   share sequential files).";
    case 0152: return "\"SIZ field invalid.";
    case 0153: return "stack too big for save area.";
    case 0154: return "system directive error.";
    case 0155: return "index tree error.";
    case 0156: return "error in file type extension on FNS too big.";
    case 0157: return "invalid user buffer addr (0, odd, or if  BLK-IO not word aligned).";
    case 0160: return "invalid user buffer size (USZ=0).";
    case 0161: return "error in version number.";
    case 0162: return "invalid volume number.";
    case 0163: return "file write error (STV=sys err code).";
    case 0164: return "device is write locked.";
    case 0165: return "error while writing prologue.";
    case 0166: return "not a valid XAB (@XAB=ODD,STV=XAB indicator).";
    case 0167: return "default directory invalid.";
    case 0170: return "cannot access argument list.";
    case 0171: return "cannot close file.";
    case 0172: return "cannot deliver AST.";
    case 0173: return "channel assignment failure (STV=sys err code).";
    case 0174: return "terminal output ignored due to (CNTRL) O.";
    case 0175: return "terminal input aborted due to (CNTRL) Y.";
    case 0176: return "default filename string address error.";
    case 0177: return "invalid device I.D.  field.";
    case 0200: return "expanded string address error.";
    case 0201: return "filename string address error.";
    case 0202: return "FSZ field invalid.";
    case 0203: return "invalid argument list.";
    case 0204: return "known file found.";
    case 0205: return "logical name error.";
    case 0206: return "node name error.";
    case 0207: return "operation successful.";
    case 0210: return "record inserted had duplicate key.";
    case 0211: return "index update error occurred-record inserted.";
    case 0212: return "record locked but read anyway.";
    case 0213: return "record  inserted  in  primary  o.k.; may  not  be accessible by secondary keys or RFA.";
    case 0214: return "file was created, but not opened.";
    case 0215: return "bad prompt buffer address.";
    case 0216: return "async.  operation pending completion.";
    case 0217: return "quoted string error.";
    case 0220: return "record header buffer invalid.";
    case 0221: return "invalid related file.";
    case 0222: return "invalid resultant string size.";
    case 0223: return "invalid resultant string address.";
    case 0224: return "operation not sequential.";
    case 0225: return "operation successful.";
    case 0226: return "created file superseded existing version.";
    case 0227: return "filename syntax error.";
    case 0230: return "time-out period expired.";
    case 0231: return "FB$BLK record attribute not supported.";
    case 0232: return "bad byte size.";
    case 0233: return "cannot disconnect RAB.";
    case 0234: return "cannot get JFN for file.";
    case 0235: return "cannot open file.";
    case 0236: return "bad JFN value.";
    case 0237: return "cannot position to end-of-file.";
    case 0240: return "cannot truncate file.";
    case 0241: return "file is currently in an  undefined  state; access is denied.";
    case 0242: return "file must be opened for exclusive access.";
    case 0243: return "directory full.";
    case 0244: return "handler not in system.";
    case 0245: return "fatal hardware error.";
    case 0246: return "attempt to write beyond EOF.";
    case 0247: return "hardware option not present.";
    case 0250: return "device not attached.";
    case 0251: return "device already attached.";
    case 0252: return "device not attachable.";
    case 0253: return "sharable resource in use.";
    case 0254: return "illegal overlay request.";
    case 0255: return "block check or CRC error.";
    case 0256: return "caller's nodes exhausted.";
    case 0257: return "index file full.";
    case 0260: return "file header full.";
    case 0261: return "accessed for write.";
    case 0262: return "file header checksum failure.";
    case 0263: return "attribute control list error.";
    case 0264: return "file already accessed on LUN.";
    case 0265: return "bad tape format.";
    case 0266: return "illegal operation on file descriptor block.";
    case 0267: return "rename; 2 different devices.";
    case 0270: return "rename; new filename already in use.";
    case 0271: return "cannot rename old file system.";
    case 0272: return "file already open.";
    case 0273: return "parity error on device.";
    case 0274: return "end of volume detected.";
    case 0275: return "data over-run.";
    case 0276: return "bad block on device.";
    case 0277: return "end of tape detected.";
    case 0300: return "no buffer space for file.";
    case 0301: return "file exceeds allocated space -- no blks.";
    case 0302: return "specified task not installed.";
    case 0303: return "unlock error.";
    case 0304: return "no file accessed on LUN.";
    case 0305: return "send/receive failure.";
    case 0306: return "spool or submit command file failure.";
    case 0307: return "no more files.";
    case 0310: return "DAP file transfer Checksum error.";
    case 0311: return "Quota exceeded";
    case 0312: return "internal network error condition detected.";
    case 0313: return "terminal input aborted due to (CNTRL) C.";
    case 0314: return "data bucket fill size > bucket size in XAB.";
    case 0315: return "invalid expanded string length.";
    case 0316: return "illegal bucket format.";
    case 0317: return "bucket size of LAN NOT = IAN in XAB.";
    case 0320: return "index not initialized.";
    case 0321: return "illegal file attributes (corrupt file header).";
    case 0322: return "index bucket fill size > bucket size in XAB.";
    case 0323: return "key name buffer not readable or writeable in XAB.";
    case 0324: return "index bucket will not hold two keys  for  key  of reference.";
    case 0325: return "multi-buffer count invalid (negative value).";
    case 0326: return "network operation failed at remote node.";
    case 0327: return "record is already locked.";
    case 0330: return "deleted record successfully accessed.";
    case 0331: return "retrieved record exceeds specified key value.";
    case 0332: return "key XAB not filled in.";
    case 0333: return "nonexistent record successfully accessed.";
    case 0334: return "unsupported prologue version.";
    case 0335: return "illegal key-of-reference in XAB.";
    case 0336: return "invalid resultant string length.";
    case 0337: return "error updating rrv's, some paths to data  may  be lost.";
    case 0340: return "data types  other  than  string  limited  to  one segment in XAB.";
    case 0341: return "reserved";
    case 0342: return "operation not supported over network.";
    case 0343: return "error on write behind.";
    case 0344: return "invalid wildcard operation.";
    case 0345: return "working set full (can not lock buffers in working set.)";
    case 0346: return "directory listing -- error in reading  volume-set name, directory name, of file name.";
    case 0347: return "directory  listing  --  error  in  reading   file attributes.";
    case 0350: return "directory  listing  --  protection  violation  in trying  to read the volume-set, directory or file name.";
    case 0351: return "directory  listing  --  protection  violation  in trying to read file attributes.";
    case 0352: return "directory  listing  --  file  attributes  do  not exist.";
    case 0353: return "directory listing -- unable to recover  directory list after Continue Transfer (Skip).";
    case 0354: return "sharing not enabled.";
    case 0355: return "sharing page count exceeded.";
    case 0356: return "UPI bit not set when sharing with BRO set.";
    case 0357: return "error in access control string (poor man's  route through error).";
    case 0360: return "terminator not seen.";
    case 0361: return "bad escape sequence.";
    case 0362: return "partial escape sequence.";
    case 0363: return "invalid wildcard context value.";
    case 0364: return "invalid directory rename operation.";
    case 0365: return "user structure (FAB/RAB)  became  invalid  during operation.";
    case 0366: return "network file transfer made precludes operation.";
    }
    sprintf(errstring, "unknown error: 0%o", code);
    return errstring;
}
